@@ -11,8 +11,13 @@ import sys
 import SCons.Script.Main
 
 
-def get_build_wheel_requires(settings):
-    return []
+# optional hooks
+#
+# def get_build_wheel_requires(settings):
+#     return []
+#
+# def get_build_sdist_requires(settings):
+#     return []
 
 
 def _run(alias):
@@ -35,10 +40,6 @@ def build_wheel(wheel_directory, settings, metadata_directory=None):
     return _run("bdist_wheel")
 
 
-def get_build_sdist_requires(settings):
-    return []
-
-
 def build_sdist(sdist_directory, settings):
     sys.argv[1:] = ["--dist-dir=" + sdist_directory, "sdist"]
     try:
